
#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <zephyr/random/rand32.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/mqtt.h>
#include <zephyr/net/tls_credentials.h>
#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(mqtt);

#define CONFIG_MQTT_MESSAGE_BUFFER_SIZE 128
#define CONFIG_MQTT_PAYLOAD_BUFFER_SIZE 128
#define MQTT_TLS_SEC_TAG                24
#define CONFIG_MQTT_BROKER_HOSTNAME     "test.mosquitto.org"
#define CONFIG_MQTT_BROKER_ADDR         "91.121.93.94"
#define CONFIG_MQTT_BROKER_PORT         8884
#define CONFIG_MQTT_SUB_TOPIC           "sensors"
#define CONFIG_MQTT_PUB_TOPIC           "sensors"
#define CONFIG_MQTT_CLIENT_ID           "zephyrTest"

static bool connected;
static uint8_t rx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static uint8_t tx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static uint8_t payload_buf[CONFIG_MQTT_PAYLOAD_BUFFER_SIZE];

static const unsigned char ca_certificate[] = {
#include "mosquitto_ca_cert.h"
};

static const unsigned char client_certificate[] = {
#include "mosquitto_client_cert.h"
};

static const unsigned char priv_key[] = {
#include "mosquitto_priv_key.h"
};

static struct sockaddr_storage broker_server;
static struct mqtt_client client;
static struct pollfd fds;


static int get_received_payload(struct mqtt_client *c, size_t length)
{
	int ret;
	int err = 0;

	if (length > sizeof(payload_buf))
    {
		err = -EMSGSIZE;
	}

	/* Truncate payload until it fits in the payload buffer. */
	while (length > sizeof(payload_buf))
    {
		ret = mqtt_read_publish_payload_blocking(c, payload_buf, (length - sizeof(payload_buf)));
		if (ret == 0)
        {
			return -EIO;
		}
        else if (ret < 0)
        {
			return ret;
		}
		length -= ret;
	}

	ret = mqtt_readall_publish_payload(c, payload_buf, length);
	if (ret)
    {
		return ret;
	}
	return err;
}

static int subscribe(struct mqtt_client *const c)
{
	struct mqtt_topic subscribe_topic = {
		.topic = {.utf8 = CONFIG_MQTT_SUB_TOPIC, .size = strlen(CONFIG_MQTT_SUB_TOPIC)},
		.qos = MQTT_QOS_1_AT_LEAST_ONCE};

	const struct mqtt_subscription_list subscription_list = {
		.list = &subscribe_topic, .list_count = 1, .message_id = 1234};

	LOG_INF("Subscribing to %s", CONFIG_MQTT_SUB_TOPIC);
	return mqtt_subscribe(c, &subscription_list);
}

static void data_print(uint8_t *prefix, uint8_t *data, size_t len)
{
	char buf[len + 1];

	memcpy(buf, data, len);
	buf[len] = 0;
	LOG_INF("%s%s", (char *)prefix, (char *)buf);
}

int publish(struct mqtt_client *c, enum mqtt_qos qos, uint8_t *data, size_t len)
{
	struct mqtt_publish_param param;

	param.message.topic.qos = qos;
	param.message.topic.topic.utf8 = CONFIG_MQTT_PUB_TOPIC;
	param.message.topic.topic.size = strlen(CONFIG_MQTT_PUB_TOPIC);
	param.message.payload.data = data;
	param.message.payload.len = len;
	param.message_id = sys_rand32_get();
	param.dup_flag = 0;
	param.retain_flag = 0;

	data_print("Publishing: ", data, len);
	LOG_INF("to topic: %s len: %u", CONFIG_MQTT_PUB_TOPIC, (unsigned int)strlen(CONFIG_MQTT_PUB_TOPIC));

	return mqtt_publish(c, &param);
}

void mqtt_evt_handler(struct mqtt_client *const c, const struct mqtt_evt *evt)
{
	int err;

	switch (evt->type)
    {
        case MQTT_EVT_CONNACK:
            if (evt->result != 0)
            {
                LOG_ERR("MQTT connect failed: %d", evt->result);
                break;
            }

            LOG_INF("MQTT client connected");
            subscribe(c);
            break;

        case MQTT_EVT_DISCONNECT:
            LOG_INF("MQTT client disconnected: %d", evt->result);
            break;

        case MQTT_EVT_PUBLISH:
            const struct mqtt_publish_param *p = &evt->param.publish;
            err = get_received_payload(c, p->message.payload.len);
            if (p->message.topic.qos == MQTT_QOS_1_AT_LEAST_ONCE)
            {
                const struct mqtt_puback_param ack = {.message_id = p->message_id};
                mqtt_publish_qos1_ack(c, &ack);
            }

            if (err >= 0)
            {
                data_print("Received: ", payload_buf, p->message.payload.len);

            }
            else if (err == -EMSGSIZE)
            {
                LOG_ERR("Received payload (%d bytes) is larger than the payload buffer size (%d bytes).", p->message.payload.len, sizeof(payload_buf));
            }
            else
            {
                LOG_ERR("get_received_payload failed: %d", err);
                LOG_INF("Disconnecting MQTT client...");

                err = mqtt_disconnect(c);
                if (err)
                {
                    LOG_ERR("Could not disconnect: %d", err);
                }
            }
            break;
        case MQTT_EVT_PUBACK:
            if (evt->result != 0)
            {
                LOG_ERR("MQTT PUBACK error: %d", evt->result);
                break;
            }
            LOG_INF("PUBACK packet id: %u", evt->param.puback.message_id);
            break;
        case MQTT_EVT_SUBACK:
            if (evt->result != 0)
            {
                LOG_ERR("MQTT SUBACK error: %d", evt->result);
                break;
            }
            LOG_INF("SUBACK packet id: %u", evt->param.suback.message_id);
            break;
        case MQTT_EVT_PINGRESP:
            if (evt->result != 0)
            {
                LOG_ERR("MQTT PINGRESP error: %d", evt->result);
            }
            break;
        default:
            LOG_INF("Unhandled MQTT event type: %d", evt->type);
            break;
	}
}

static void broker_init(void)
{
	struct sockaddr_in *broker4 = (struct sockaddr_in *)&broker_server;

	broker4->sin_family = AF_INET;
	broker4->sin_port = htons(CONFIG_MQTT_BROKER_PORT);
	zsock_inet_pton(AF_INET, CONFIG_MQTT_BROKER_ADDR, &broker4->sin_addr);
}

int client_init(struct mqtt_client *client)
{
	int err;
	mqtt_client_init(client);
    broker_init();

	client->broker = &broker_server;
	client->evt_cb = mqtt_evt_handler;
	client->client_id.utf8 = CONFIG_MQTT_CLIENT_ID;
	client->client_id.size = strlen(client->client_id.utf8);
	client->password = NULL;
	client->user_name = NULL;
	client->protocol_version = MQTT_VERSION_3_1_1;
	client->rx_buf = rx_buffer;
	client->rx_buf_size = sizeof(rx_buffer);
	client->tx_buf = tx_buffer;
	client->tx_buf_size = sizeof(tx_buffer);

	/* STEP 4 - Set the transport type to secure */
	struct mqtt_sec_config *tls_cfg = &(client->transport).tls.config;
	static sec_tag_t sec_tag_list[] = {MQTT_TLS_SEC_TAG};

	client->transport.type = MQTT_TRANSPORT_SECURE;

	tls_cfg->peer_verify = TLS_PEER_VERIFY_OPTIONAL;
	tls_cfg->cipher_count = 0;
	tls_cfg->cipher_list = NULL;
	tls_cfg->sec_tag_count = ARRAY_SIZE(sec_tag_list);
	tls_cfg->sec_tag_list = sec_tag_list;
	tls_cfg->session_cache = TLS_SESSION_CACHE_DISABLED;
	tls_cfg->hostname = CONFIG_MQTT_BROKER_HOSTNAME;

	return err;
}

int mqtt_start()
{
	int err;

	// CA Certificate
    err = tls_credential_add(MQTT_TLS_SEC_TAG, TLS_CREDENTIAL_CA_CERTIFICATE, ca_certificate, sizeof(ca_certificate));
	if (err < 0)
    {
		LOG_ERR("Failed to add TLS credentials CA Certificate, err: %d", err);
		return err;
	}

	// Client Certificate
    err = tls_credential_add(MQTT_TLS_SEC_TAG, TLS_CREDENTIAL_SERVER_CERTIFICATE, client_certificate, sizeof(client_certificate));
	if (err < 0)
    {
		LOG_ERR("Failed to add TLS credentials Client Certificate, err: %d", err);
		return err;
	}

	// Private Key
    err = tls_credential_add(MQTT_TLS_SEC_TAG, TLS_CREDENTIAL_PRIVATE_KEY, priv_key, sizeof(priv_key));
	if (err < 0)
    {
		LOG_ERR("Failed to add TLS credentials Private Key, err: %d", err);
		return err;
	}

	LOG_INF("Connecting to MQTT broker");

	err = client_init(&client);
	if (err)
    {
		LOG_ERR("Failed to initialize MQTT client: %d", err);
		return err;
	}

	err = mqtt_connect(&client);
	if (err)
    {
		LOG_ERR("Error in mqtt_connect: %d", err);
		return err;
	}

	/* STEP 5 - Update the file descriptor to use the TLS socket */
	fds.fd = client.transport.tls.sock;
	fds.events = POLLIN;

	while (1)
    {
		err = poll(&fds, 1, mqtt_keepalive_time_left(&client));
		if (err < 0) {
			LOG_ERR("Error in poll(): %d", errno);
			break;
		}

		err = mqtt_live(&client);
		if ((err != 0) && (err != -EAGAIN))
        {
			LOG_ERR("Error in mqtt_live: %d", err);
			break;
		}
		if ((fds.revents & POLLIN) == POLLIN)
        {
			err = mqtt_input(&client);
			if (err != 0)
            {
				LOG_ERR("Error in mqtt_input: %d", err);
				break;
			}
		}
		if ((fds.revents & POLLERR) == POLLERR)
        {
			LOG_ERR("POLLERR");
			break;
		}
		if ((fds.revents & POLLNVAL) == POLLNVAL)
        {
			LOG_ERR("POLLNVAL");
			break;
		}
	}

	LOG_INF("Disconnecting MQTT client");
	err = mqtt_disconnect(&client);
	if (err) {
		LOG_ERR("Could not disconnect MQTT client: %d", err);
		return err;
	}
}